import time
import sys
import os
import shutil
import random
import tempfile
import functools
import json
import requests
import math
from zipfile import ZipFile
from string import Template

import shortuuid
import boto3
import botocore

try:
    import importlib.resources as pkg_resources
except ImportError:
    import importlib_resources as pkg_resources

from . import main  # relative-import the *package* containing the templates

from .utils import *

class create_prediction_api_class():

    def __init__(self, model_filepath, unique_model_id, model_type, categorical, labels, apiid, custom_libraries, requirements, repo_name="", image_tag="", memory=None, timeout=90, pyspark_support=False):

        #####
        self.user_session = boto3.session.Session(
            aws_access_key_id=os.environ.get("AWS_ACCESS_KEY_ID_AIMS"),
            aws_secret_access_key = os.environ.get("AWS_SECRET_ACCESS_KEY_AIMS"), 
            region_name=os.environ.get("AWS_REGION_AIMS")
        )
        self.sts_client = self.user_session.client("sts")
        self.account_id = self.sts_client.get_caller_identity()["Account"]
        #####

        from .aws_client import AWSClient
        self.aws_client = AWSClient(self.user_session)
                
        #####
        self.account_id = self.account_id
        self.unique_model_id = unique_model_id
        self.categorical = categorical
        self.model_type = model_type
        self.labels = labels
        self.apiid = apiid
        self.custom_libraries = custom_libraries
        self.requirements = requirements
        self.repo_name = repo_name
        self.image_tag = image_tag
        self.memory = memory
        self.timeout = timeout
        self.pyspark_support = pyspark_support
        self.region = os.environ.get("AWS_REGION")
        self.bucket_name = os.environ.get("BUCKET_NAME")
        self.python_runtime = os.environ.get("PYTHON_RUNTIME", "python3.12")
        #####

        self.model_type = self.model_type.lower()

        if(self.categorical == "TRUE"):
            self.categorical = True
        elif(self.categorical == "FALSE"):
            self.categorical = False

        if(self.custom_libraries == "TRUE"):
            self.custom_libraries = True
        elif(self.custom_libraries == "FALSE"):
            self.custom_libraries = False

        self.memory_model_mapping = {
            "tabular": 1024,
            "text": 1024,
            "image": 1024,
            "video": 1024,
            "custom":1024
        }

        self.timeout_model_mapping = {
            "tabular": 90,
            "text": 90,
            "image": 90,
            "video": 90,
            "custom": 90
        }
        # UPDATED: New eval layer ARNs (python3.12)
        self.eval_layer_map = {
            "us-east-1": "arn:aws:lambda:us-east-1:585666012274:layer:eval-layer-python3-12:4",
            "us-east-2": "arn:aws:lambda:us-east-2:517169013426:layer:eval_layer_test:5",
            "us-west-1": "arn:aws:lambda:us-west-1:517169013426:layer:eval_layer_test:1",
            "us-west-2": "arn:aws:lambda:us-west-2:517169013426:layer:eval_layer_test:1",
            "eu-west-1": "arn:aws:lambda:eu-west-1:585666012274:layer:eval-layer-python3-12:1",
            "eu-west-2": "arn:aws:lambda:eu-west-2:517169013426:layer:eval_layer_test:1",
            "eu-west-3": "arn:aws:lambda:eu-west-3:517169013426:layer:eval_layer_test:1"
        }

        # UPDATED: New auth layer ARNs (python3.12)
        self.auth_layer_map = {
            "us-east-1": "arn:aws:lambda:us-east-1:585666012274:layer:aimsauth-layer-python3-12:6",
            "us-east-2": "arn:aws:lambda:us-east-2:517169013426:layer:aimsauth_layer:9",
            "us-west-1": "arn:aws:lambda:us-west-1:517169013426:layer:aimsauth_layer:1",
            "us-west-2": "arn:aws:lambda:us-west-2:517169013426:layer:aimsauth_layer:1",
            "eu-west-1": "arn:aws:lambda:eu-west-1:585666012274:layer:aimsauth-layer-python3-12:6",
            "eu-west-2": "arn:aws:lambda:eu-west-2:517169013426:layer:aimsauth_layer:1",
            "eu-west-3": "arn:aws:lambda:eu-west-3:517169013426:layer:aimsauth_layer:1"
        }
        
        try:
            onnx_size = math.ceil(os.path.getsize(model_filepath)/(1024*1024))
        except:
            onxx_size = 500

        self.temp_dir_file_deletion_list = ['archive2.zip', 'archive3.zip', 'archive.zip', 'archivetest.zip', 'archiveeval.zip', 'archiveauth.zip', 'main.py', 'ytest.pkl']
        self.memory = self.memory_model_mapping[self.model_type] if self.memory==None else memory
        self.timeout = self.timeout_model_mapping[self.model_type] if self.timeout==None else timeout

        self.eval_layer = self.eval_layer_map[self.region]
        self.auth_layer = self.auth_layer_map[self.region]

        self.temp_dir = tempfile.gettempdir()
        self.file_objects_folder_path = os.path.join(self.temp_dir, 'file_objects')

        self.memory_lambda = {
            "main": 1024,
            "eval": 2048,
            "auth": 512
        }

        if self.categorical == True:
            self.task_type="classification"
        elif self.categorical == False:
            self.task_type="regression"
        else:
            self.task_type="custom"
        
        if(self.model_type=="custom"):
            self.model_class="custom"
        elif(self.model_type=="neural style transfer"):
            self.model_class="generative"
        elif(self.categorical==True):
            self.model_class="classification"
        elif(self.categorical==False):
            self.model_class="regression"

        self.model_template_mapping = {
            "classification": {
                "text": "1.txt",
                "image": "2.txt",
                "tabular": "4.txt",
                "audio": "7.txt",
                "video": "8.txt"
            },
            "regression": {
                "text": "1B.txt",
                "image": "3.txt",
                "tabular": "5.txt",
                "timeseries": "6.txt"
            },
            "generative": {
                "neural style transfer": "nst.txt"
            }
        }

    def create_prediction_api(self):
        global main
        if self.pyspark_support:
            try:
                from . import pyspark as main
            except:
                raise("Error: Please install pyspark to enable pyspark features")
        
        delete_files_from_temp_dir(self.temp_dir_file_deletion_list)

        if self.model_type != "custom":
            delete_folder(self.file_objects_folder_path)
            make_folder(self.file_objects_folder_path)

        if self.model_type == 'custom':
            with open("custom_lambda.py", 'r') as in_file:     
                newdata = in_file.read()
        else:
            data = pkg_resources.read_text(main, self.model_template_mapping[self.model_class][self.model_type])
            t = Template(data)
            if(self.model_class=="classification"):
                newdata = t.substitute(
                    bucket_name=os.environ.get("BUCKET_NAME"),
                    unique_model_id=self.unique_model_id,
                    labels=self.labels
                )
            else:
                newdata = t.substitute(
                    bucket_name=os.environ.get("BUCKET_NAME"),
                    unique_model_id=self.unique_model_id
                )

        with open(os.path.join(self.file_objects_folder_path, 'model.py'), 'w') as file:
            file.write(newdata)

        if(self.model_type == 'custom'):
            from . import custom_approach
            data = pkg_resources.read_text(custom_approach, 'lambda_function.py')
        else:
            data = pkg_resources.read_text(main, 'lambda_function.txt')

        with open(os.path.join(self.file_objects_folder_path, 'lambda_function.py'), 'w') as file:
            file.write(data)
        
        ###
        api_key = str(shortuuid.uuid())

        t = Template(pkg_resources.read_text(main, 'eval_lambda.txt').replace("$apikey",api_key).replace("$task_type",self.task_type))
        
        data = t.substitute(bucket_name = self.bucket_name, unique_model_id = self.unique_model_id, task_type = self.task_type)
        with open(os.path.join(self.temp_dir, 'main.py'), 'w') as file:
            file.write(data)
        with ZipFile(os.path.join(self.temp_dir, 'archive2.zip'), 'a') as z:
            z.write(os.path.join(self.temp_dir, 'main.py'), 'main.py')
        self.aws_client.upload_file_to_s3(os.path.join(self.temp_dir, 'archive2.zip'), os.environ.get("BUCKET_NAME"), self.unique_model_id+"/"+'archiveeval.zip')

        data2 = pkg_resources.read_text(main, 'authorization.txt').replace("$apikey",api_key)
        with open(os.path.join(self.temp_dir, 'main.py'), 'w') as file:
            file.write(data2)
        with ZipFile(os.path.join(self.temp_dir, 'archive3.zip'), 'a') as z:
            z.write(os.path.join(self.temp_dir, 'main.py'), 'main.py')
        self.aws_client.upload_file_to_s3(os.path.join(self.temp_dir, 'archive3.zip'), os.environ.get("BUCKET_NAME"), self.unique_model_id+"/"+'archiveauth.zip')
        ###

        if self.model_type.lower() == 'custom':
            self.aws_client.upload_file_to_s3(os.path.join(self.temp_dir, 'exampledata.json'), os.environ.get("BUCKET_NAME"), self.unique_model_id+"/"+"exampledata.json")
    
        delete_files_from_temp_dir(self.temp_dir_file_deletion_list)

        ####################

        short_uuid = str(shortuuid.uuid())

        lambdarolename = 'myService-dev-us-' + self.region + '-lambdaRole'+short_uuid
        lambdapolicyname = 'myService-dev-' + self.region + '-lambdaPolicy'+short_uuid
        lambdafxnname = 'modfunction'+short_uuid
        lambdaauthfxnname = 'redisAccess'+short_uuid
        lambdaevalfxnname = 'evalfunction'+short_uuid

        from . import json_templates

        lambdarole1 = json.loads(pkg_resources.read_text(json_templates, 'lambda_role_1.txt'))
        lambdapolicy1 = json.loads(pkg_resources.read_text(json_templates, 'lambda_policy_1.txt'))

        self.aws_client.delete_iam_role(lambdarolename) # delete role for CodeBuild if role with same name exists
        self.aws_client.create_iam_role(lambdarolename, lambdarole1) # creating role for CodeBuild
        self.aws_client.delete_iam_policy(lambdapolicyname) # delete policy for CodeBuild if policy with same name exists
        self.aws_client.create_iam_policy(lambdapolicyname, lambdapolicy1) # creating policy for CodeBuild
        self.aws_client.attach_policy_to_role(lambdarolename, lambdapolicyname)

        sys.stdout.write('\r')
        sys.stdout.write("[============                         ] Progress: 40% - Creating custom containers...                        ")
        sys.stdout.flush()

        #########

        if(self.custom_libraries == False):
            from aimodelshare.containerization import create_lambda_using_base_image
            response6 = create_lambda_using_base_image(self.user_session, os.getenv("BUCKET_NAME"), self.file_objects_folder_path, lambdafxnname, self.apiid, self.repo_name, self.image_tag, self.memory, self.timeout)
        elif(self.custom_libraries == True):
            requirements = self.requirements.split(",")
            for i in range(len(requirements)):
                requirements[i] = requirements[i].strip(" ")
            with open(os.path.join(self.file_objects_folder_path, 'requirements.txt'), 'a') as f:
                for lib in requirements:
                    f.write('%s\n' % lib)
            requirements_file_path = os.path.join(self.file_objects_folder_path, 'requirements.txt')
            from aimodelshare.containerisation import deploy_container
            response6 = deploy_container(self.account_id, os.environ.get("AWS_REGION"), self.user_session, lambdafxnname, self.file_objects_folder_path,requirements_file_path,self.apiid, pyspark_support=self.pyspark_support)

        ##########

        lambdaclient = self.user_session.client('lambda')
        role_arn = 'arn:aws:iam::' + self.account_id + ':role/' + lambdarolename
        handler = 'main.handler'

        def create_lambda_function(function_name, python_runtime, role_arn, handler, code_source, timeout, memory_size, layers):
            response = self.aws_client.lambda_client.create_function(
                FunctionName = function_name,
                Runtime = python_runtime,
                Role = role_arn,
                Handler = handler,
                Code = code_source,
                Timeout = timeout,
                MemorySize = memory_size,
                Layers = layers
            )

        eval_code_source = {'S3Bucket': self.bucket_name, 'S3Key':  self.unique_model_id + "/" + "archiveeval.zip"}
        eval_layers = [self.eval_layer, self.auth_layer]
        create_lambda_function(lambdaevalfxnname, self.python_runtime, role_arn, handler, eval_code_source, 90, 2048, eval_layers)

        auth_code_source = {'S3Bucket': self.bucket_name, 'S3Key':  self.unique_model_id + "/" + "archiveauth.zip"}
        auth_layers = [self.auth_layer]
        create_lambda_function(lambdaauthfxnname, self.python_runtime, role_arn, handler, auth_code_source, 90, 512, auth_layers)

        sys.stdout.write('\r')
        sys.stdout.write("[==========================           ] Progress: 75% - Deploying prediction API...                          ")
        sys.stdout.flush()
        # }}}

        ##############################

        api_id = self.apiid
        stmt_id = 'apigateway-prod-'+str(shortuuid.uuid())

        resourceidlist = self.aws_client.get_api_resources(api_id)
        api_path_id = {}
        for i in resourceidlist:
            api_path_id.update({i['path']: i['id']})

        resource_id_parent = api_path_id['/']
        resource_id_lambda = api_path_id['/m']
        resource_id_eval = api_path_id['/eval']

        fxn_list = self.aws_client.lambda_client.list_functions()

        arn_prefix = "arn:aws:execute-api:" + self.region + ":" + self.account_id + ":" + api_id
        self.aws_client.add_invoke_resource_policy_to_lambda(lambdaauthfxnname, stmt_id, arn_prefix + "/*/*")
        #self.aws_client.add_invoke_resource_policy_to_lambda(lambdafxnname, 'apigateway-prod-2', arn_prefix + "/*/POST/m")
        #self.aws_client.add_invoke_resource_policy_to_lambda(lambdafxnname, 'apigateway-test-2', arn_prefix + "/*/POST/m")
        self.aws_client.add_invoke_resource_policy_to_lambda(lambdaevalfxnname, 'apigateway-prod-3', arn_prefix + "/*/POST/eval")
        self.aws_client.add_invoke_resource_policy_to_lambda(lambdaevalfxnname, 'apigateway-test-3', arn_prefix + "/*/POST/eval")

        integration_response = json.loads(pkg_resources.read_text(json_templates, 'integration_response.txt'))

        lambdarole2 = json.loads(pkg_resources.read_text(json_templates, 'lambda_role_2.txt'))
        lambdarolename2 = 'lambda_invoke_function_assume_apigw_role_2'
        lambdapolicy2 = json.loads(pkg_resources.read_text(json_templates, 'lambda_policy_2.txt'))
        lambdapolicyname2 = 'invokelambda'

        self.aws_client.delete_iam_role(lambdarolename2) # delete role for CodeBuild if role with same name exists
        self.aws_client.create_iam_role(lambdarolename2, lambdarole2) # creating role for CodeBuild
        self.aws_client.delete_iam_policy(lambdapolicyname2) # delete policy for CodeBuild if policy with same name exists
        self.aws_client.create_iam_policy(lambdapolicyname2, lambdapolicy2) # creating policy for CodeBuild
        self.aws_client.attach_policy_to_role(lambdarolename2, lambdapolicyname2)

        uri_str = "arn:aws:apigateway:" + self.region + ":lambda:path/2015-03-31/functions/arn:aws:lambda:" + self.region + ":" + self.account_id + ':function:' + lambdafxnname + '/invocations'
        credentials = 'arn:aws:iam::'+self.account_id+':role/' + lambdarolename2
        self.aws_client.integration_setup(api_id, resource_id_lambda, uri_str, credentials, integration_response)

        uri_str_2 = "arn:aws:apigateway:" + self.region + ":lambda:path/2015-03-31/functions/arn:aws:lambda:" + self.region + ":" + self.account_id + ':function:' + lambdaevalfxnname + '/invocations'
        credentials_2 = 'arn:aws:iam::'+self.account_id+':role/' + lambdarolename2
        self.aws_client.integration_setup(api_id, resource_id_eval, uri_str_2, credentials_2, integration_response)

        response = self.aws_client.apigateway_client.update_rest_api(
            restApiId=api_id,
            patchOperations=[
                {
                    "op": "replace",
                    "path": "/policy",
                    "value": '{"Version": "2012-10-17","Statement": [{"Effect": "Allow","Principal": "*","Action": "execute-api:Invoke","Resource": "arn:aws:execute-api:'+os.environ.get("AWS_REGION")+':'+self.account_id+':'+api_id+'/prod/OPTIONS/*"}]}'
                }
            ]
        )

        auth_uri_str = "arn:aws:apigateway:"+ os.environ.get("AWS_REGION") +":lambda:path/2015-03-31/functions/arn:aws:lambda:"+os.environ.get("AWS_REGION")+":"+self.account_id+":function:"+lambdaauthfxnname+"/invocations"

        responseauthfxnapigateway = self.aws_client.apigateway_client.create_authorizer(
            restApiId=api_id,
            name='aimscustomauthfxn',
            type='TOKEN',
            authorizerUri=auth_uri_str,
            identitySource="method.request.header.authorizationToken",
            authorizerResultTtlInSeconds=0
        )

        responseauthfxnapigateway = self.aws_client.apigateway_client.get_authorizers(
            restApiId=api_id
        )

        authorizerid=responseauthfxnapigateway['items'][0]['id']

        stmt_idauth = 'apigateway-prod-'+str(shortuuid.uuid())
        response70 = self.user_session.client('lambda').add_permission(
            FunctionName=lambdaauthfxnname,
            StatementId=stmt_idauth,
            Action='lambda:InvokeFunction',
            Principal='apigateway.amazonaws.com',
            SourceArn='arn:aws:execute-api:'+self.region+':' + self.account_id+':'+api_id + '/authorizers/' + authorizerid
        )
        
        response_modmthd_addauth = self.user_session.client('apigateway').update_method(
            restApiId=api_id,
            resourceId=resource_id_lambda,
            httpMethod='POST',
            patchOperations=[
                {
                    'op': 'replace',
                    'path': '/authorizationType',
                    'value': 'CUSTOM',
                    'from': 'NONE'
                },
                {
                    'op': 'replace',
                    'path': '/authorizerId',
                    'value': authorizerid
                }
            ]
        )

        response_evalmthd_addauth = self.user_session.client('apigateway').update_method(
            restApiId=api_id,
            resourceId=resource_id_eval,
            httpMethod='POST',
            patchOperations=[
                {
                    'op': 'replace',
                    'path': '/authorizationType',
                    'value': 'CUSTOM',
                    'from': 'NONE'
                },
                {
                    'op': 'replace',
                    'path': '/authorizerId',
                    'value': authorizerid
                }])
        response12 = self.user_session.client('apigateway').create_deployment(
            restApiId=api_id,
            stageName='prod'
        )

        result = 'https://'+ api_id + '.execute-api.' + os.environ.get("AWS_REGION") + '.amazonaws.com/prod/m'

        if self.model_type=='custom':
            ### Progress Update #6/6 {{{
            sys.stdout.write('\r')
            sys.stdout.write("[=====================================] Progress: 100% - API deployment completed!                          ")
            sys.stdout.flush()
            # }}}

        return {"statusCode": 200,
            "headers": {
                "Access-Control-Allow-Origin": "*",
                "Access-Control-Allow-Credentials": True
            },
            "body": json.dumps(result)}



def create_prediction_api(model_filepath, unique_model_id, model_type, 
                          categorical, labels, apiid, custom_libraries, 
                          requirements, repo_name="", image_tag="", 
                          memory=None, timeout=None, pyspark_support=False):
    api_class = create_prediction_api_class(model_filepath, unique_model_id, 
                                           model_type, categorical, labels, apiid, 
                                           custom_libraries, requirements, repo_name, 
                                           image_tag, memory, timeout, pyspark_support)
    return api_class.create_prediction_api()

def get_api_json():
    region = "us-east-2"
    apijson = '''
    {
      "openapi": "3.0.1",
      "info": {
        "title": "modapi36146",
        "description": "This is a copy of my first API okay",
        "version": "2020-05-12T21:25:38Z"
      },
      "servers": [
        {
          "url": "https://8nee9nskdb.execute-api.%s.amazonaws.com/{basePath}",
          "variables": {
            "basePath": {
              "default": "/prod"
            }
          }
        }
      ],
      "paths": {
        "/eval": {
          "post": {
            "responses": {
              "200": {
                "description": "200 response",
                "headers": {
                  "Access-Control-Allow-Origin": {
                    "schema": {
                      "type": "string"
                    }
                  }
                },
                "content": {
                  "application/json": {
                    "schema": {
                      "$ref": "#/components/schemas/outputmodel"
                    }
                  }
                }
              }
            }
          },
          "options": {
            "responses": {
              "200": {
                "description": "200 response",
                "headers": {
                  "Access-Control-Allow-Origin": {
                    "schema": {
                      "type": "string"
                    }
                  },
                  "Access-Control-Allow-Methods": {
                    "schema": {
                      "type": "string"
                    }
                  },
                  "Access-Control-Allow-Headers": {
                    "schema": {
                      "type": "string"
                    }
                  }
                },
                "content": {
                  "application/json": {
                    "schema": {
                      "$ref": "#/components/schemas/Empty"
                    }
                  }
                }
              }
            }
          }
        },
        "/m": {
          "post": {
            "responses": {
              "200": {
                "description": "200 response",
                "headers": {
                  "Access-Control-Allow-Origin": {
                    "schema": {
                      "type": "string"
                    }
                  }
                },
                "content": {
                  "application/json": {
                    "schema": {
                      "$ref": "#/components/schemas/outputmodel"
                    }
                  }
                }
              }
            }
          },
          "options": {
            "responses": {
              "200": {
                "description": "200 response",
                "headers": {
                  "Access-Control-Allow-Origin": {
                    "schema": {
                      "type": "string"
                    }
                  },
                  "Access-Control-Allow-Methods": {
                    "schema": {
                      "type": "string"
                    }
                  },
                  "Access-Control-Allow-Headers": {
                    "schema": {
                      "type": "string"
                    }
                  }
                },
                "content": {
                  "application/json": {
                    "schema": {
                      "$ref": "#/components/schemas/Empty"
                    }
                  }
                }
              }
            }
          }
        }
      },
      "components": {
        "schemas": {
          "Empty": {
            "title": "Empty Schema",
            "type": "object"
          },
          "outputmodel": {
            "title": "Output",
            "type": "object",
            "properties": {
              "body": {
                "type": "string"
              }
            }
          }
        }
      },
      "x-amazon-apigateway-policy": {
        "Version": "2012-10-17",
        "Statement": [
          {
            "Effect": "Allow",
            "Principal": "*",
            "Action": "execute-api:Invoke",
            "Resource": "arn:aws:execute-api:%s:517169013426:8nee9nskdb/prod/OPTIONS/*"
          }
        ]
      }
    }
            ''' % (region, region)
    return apijson

def delete_deployment(apiurl, confirmation=True):
    """
    apiurl: string of API URL the user wishes to delete
    WARNING: User must supply high-level credentials in order to delete an API. 
    """
    from aimodelshare.aws import run_function_on_lambda

    if confirmation==True:
        # Provide Warning & Have user confirm deletion 
        print("Running this function will permanently delete all resources tied to this deployment, \n including the eval lambda and all models submitted to the model competition.\n")
        confirmation = input(prompt="To confirm, type 'permanently delete':")
        if confirmation.lower() == "permanently delete" or confirmation.lower() == "'permanently delete'":
            pass
        else:
            return print("'Delete Deployment' unsuccessful: operation cancelled by user.")
    else:
      pass
    import os 
    if os.environ.get("cloud_location") is not None:
        cloudlocation=os.environ.get("cloud_location")
    else:
        cloudlocation="not set"
    if "model_share"==cloudlocation:
            def nonecheck(objinput=""):
                if objinput==None:
                  objinput="None"
                else:
                  objinput="'/tmp/"+objinput+"'"
                return objinput

            delplaygroundstring="delete_deployment('"+apiurl+"',"+",confirmation=False)"
            import base64
            import requests
            import json

            api_url = "https://z4kvag4sxdnv2mvs2b6c4thzj40bxnuw.lambda-url.us-east-2.on.aws/"

            data = json.dumps({"code": """from aimodelshare.api import delete_deployment;"""+delplaygroundstring, "zipfilename": "","username":os.environ.get("username"), "password":os.environ.get("password"),"token":os.environ.get("JWT_AUTHORIZATION_TOKEN"),"s3keyid":"xrjpv1i7xe"})

            headers = {"Content-Type": "application/json"}
            response = requests.request("POST", api_url, headers = headers, data=data)
            # Print response
            result=json.loads(response.text)

            for i in json.loads(result['body']):
                print(i)
    
    else:

            # Confirm that creds are loaded, print warning if not
            if all(["AWS_ACCESS_KEY_ID_AIMS" in os.environ, 
                    "AWS_SECRET_ACCESS_KEY_AIMS" in os.environ,
                    "AWS_REGION_AIMS" in os.environ,
                  "username" in os.environ, 
                  "password" in os.environ]):
                pass
            else:
                return print("'Delete Deployment' unsuccessful. Please provide credentials with set_credentials().")
            
            # get api_id from apiurl
            api_url_trim = apiurl.split('https://')[1]
            api_id = api_url_trim.split(".")[0]

            # Create User Session
            user_sess = boto3.session.Session(aws_access_key_id=os.environ.get('AWS_ACCESS_KEY_ID_AIMS'), 
                                              aws_secret_access_key=os.environ.get('AWS_SECRET_ACCESS_KEY_AIMS'), 
                                              region_name=os.environ.get('AWS_REGION_AIMS'))
            
            s3 = user_sess.resource('s3')

            # Get bucket and model_id subfolder for user based on apiurl {{{
            response, error = run_function_on_lambda(
                apiurl, **{"delete": "FALSE", "versionupdateget": "TRUE"}
            )
            if error is not None:
                raise error
            import json
            _, api_bucket, model_id = json.loads(response.content.decode("utf-8"))
            # }}} 

            #Confirm username in bucket name
            ## TODO: Update this check to more secure process
            if os.environ.get("username").lower() in api_bucket:
                pass
            else:
                print("Permission denied. Please provide credentials that allow administrator access to this api.")
                return


            # get api resources
            api = user_sess.client('apigateway')
            resources = api.get_resources(
                restApiId=api_id
                )

            #get lambda arns 
            lambda_arns = list()
            for i in range(len(resources['items'])):
                if len(resources['items'][i]) > 2: 
                    resource_id = resources['items'][i]['id']
                    integration = api.get_integration(
                        restApiId=api_id,
                        resourceId=resource_id,
                        httpMethod='POST'
                    )
                    uri=integration['uri']
                    ans1=uri.split('functions/')
                    lambda_arn = ans1[1].split('/invocations')[0]
                    lambda_arns.append(lambda_arn)
            else: 
                pass 

            # get authorizer arn 
            authorizers = api.get_authorizers(
                          restApiId=api_id
                          )

            authorizer_full_arn=authorizers['items'][0]['authorizerUri']
            ans1=authorizer_full_arn.split('functions/')
            auth_arn = ans1[1].split('/invocations')[0]
            lambda_arns.append(auth_arn)

            # delete lambdas & authorizer
            client = boto3.client('lambda', region_name=os.environ.get("AWS_REGION"))
            for arn in lambda_arns:
                lambda_response = client.delete_function(
                    FunctionName = arn
                )

            # delete api
            client = boto3.client('apigateway', region_name=os.environ.get("AWS_REGION"))
            api_response = client.delete_rest_api(
                restApiId=api_id
            )
            import requests

            # delete api page on front end
            bodydata = {'apiurl': apiurl,
                        'delete': "TRUE"
                        }
            headers_with_authentication = {'Content-Type': 'application/json', 'authorizationToken': os.environ.get("JWT_AUTHORIZATION_TOKEN"), 'Access-Control-Allow-Headers':
                                          'Content-Type,X-Amz-Date,authorizationToken,Access-Control-Allow-Origin,X-Api-Key,X-Amz-Security-Token,Authorization', 'Access-Control-Allow-Origin': '*'}

            requests.post("https://bhrdesksak.execute-api.us-east-1.amazonaws.com/dev/modeldata",
                          json=bodydata, headers=headers_with_authentication)
            
            # delete competition posting
            bodydata = {"apiurl": apiurl,
                        "delete":"TRUE",
                        "experiment":"FALSE"
                                        }

            # Get the response
            headers_with_authentication = {'Content-Type': 'application/json', 'authorizationToken': os.environ.get("AWS_TOKEN"), 'Access-Control-Allow-Headers':
                                            'Content-Type,X-Amz-Date,authorizationToken,Access-Control-Allow-Origin,X-Api-Key,X-Amz-Security-Token,Authorization', 'Access-Control-Allow-Origin': '*'}
            # competitiondata lambda function invoked through below url to update model submissions and contributors
            requests.post("https://o35jwfakca.execute-api.us-east-1.amazonaws.com/dev/modeldata",
                          json=bodydata, headers=headers_with_authentication)

            # delete experiment posting
            bodydata = {"apiurl": apiurl,
                        "delete":"TRUE",
                        "experiment":"TRUE"
                                        }

            # Get the response
            headers_with_authentication = {'Content-Type': 'application/json', 'authorizationToken': os.environ.get("AWS_TOKEN"), 'Access-Control-Allow-Headers':
                                            'Content-Type,X-Amz-Date,authorizationToken,Access-Control-Allow-Origin,X-Api-Key,X-Amz-Security-Token,Authorization', 'Access-Control-Allow-Origin': '*'}
            # competitiondata lambda function invoked through below url to update model submissions and contributors
            requests.post("https://o35jwfakca.execute-api.us-east-1.amazonaws.com/dev/modeldata",
                          json=bodydata, headers=headers_with_authentication)    
            
            # Delete competition data container image
            try:
                    content_object = s3.Object(bucket_name=api_bucket, key=model_id + "/competitionuserdata.json")
                    file_content = content_object.get()['Body'].read().decode('utf-8')
                    json_content = json.loads(file_content)
                    ecr_uri=json_content['datauri']      

                    ecr_client = user_sess.client('ecr-public')

                    repository_image = ecr_uri.split('/')[2]

                    repository = repository_image.split(':')[0]
                    image = repository_image.split(':')[1]

                    response = ecr_client.batch_delete_image(
                        repositoryName=repository,
                        imageIds=[
                            {
                                'imageTag': image
                            }
                        ]
                    )

                    image_details = ecr_client.describe_images(
                        repositoryName=repository
                    )

                    if len(image_details['imageDetails'])==0:
                        response = ecr_client.delete_repository(
                            repositoryName=repository
                        )
            except:
                pass
            # delete s3 folder
            bucket = s3.Bucket(api_bucket)
            bucket.objects.filter(Prefix= model_id+'/').delete() 

            return "Deployment deleted successfully."


__all__ = [
    get_api_json,
    create_prediction_api,
    delete_deployment
]



#!/usr/bin/env python3

import os
import sys
import time
import uuid
import logging
import random
import json
from datetime import datetime
from typing import Optional, List, Dict, Any, Tuple
from urllib.parse import urlparse

RUN_ID = uuid.uuid4().hex[:8]
LOG_FILE = f"/tmp/mc_comprehensive_test_{RUN_ID}.log"

# Optional flag to preserve the table across runs to accumulate users/rankings
SKIP_CLEANUP = os.environ.get("MC_SKIP_CLEANUP", "true").lower() == "true"

logger = logging.getLogger("mc_comprehensive_single_user_test")
logger.setLevel(logging.INFO)
fmt = logging.Formatter("%(asctime)s [%(levelname)s] %(message)s")

sh = logging.StreamHandler(sys.stdout)
sh.setFormatter(fmt)
logger.addHandler(sh)

fh = logging.FileHandler(LOG_FILE, mode="w", encoding="utf-8")
fh.setFormatter(fmt)
logger.addHandler(fh)

def log_kv(title: str, data: dict):
    logger.info(f"--- {title} ---")
    for k, v in data.items():
        logger.info(f"{k}: {v}")
    logger.info("--- end ---")

try:
    from aimodelshare.moral_compass import MoralcompassApiClient, ApiClientError, NotFoundError
    from aimodelshare.aws import get_token_from_session, _get_username_from_token
except ImportError as e:
    logger.error(f"Failed to import required modules: {e}")
    sys.exit(1)


class MoralCompassIntegrationTest:
    # Pools to vary run data
    ACCURACY_CHOICES = [0.72, 0.80, 0.85, 0.90, 0.95]
    TEAM_CHOICES = ["team-a", "team-b", "team-c"]
    PARTIAL_RATIO_MIN = 0.30
    PARTIAL_RATIO_MAX = 0.80

    def __init__(self, api_base_url: Optional[str] = None, session_id: Optional[str] = None):
        env_snapshot = {
            "MORAL_COMPASS_API_BASE_URL": os.environ.get("MORAL_COMPASS_API_BASE_URL"),
            "SESSION_ID_present": bool(os.environ.get("SESSION_ID")),
            "JWT_AUTHORIZATION_TOKEN_present": bool(os.environ.get("JWT_AUTHORIZATION_TOKEN")),
            "TEST_TABLE_ID": os.environ.get("TEST_TABLE_ID"),
            "TEST_PLAYGROUND_URL": os.environ.get("TEST_PLAYGROUND_URL"),
            "MC_SKIP_CLEANUP": os.environ.get("MC_SKIP_CLEANUP", "true"),
            "PYTHON_VERSION": sys.version.split()[0],
            "RUN_ID": RUN_ID,
            "LOG_FILE": LOG_FILE,
        }
        log_kv("Environment Snapshot", env_snapshot)

        api_base_url = api_base_url or os.environ.get("MORAL_COMPASS_API_BASE_URL")
        if not api_base_url:
            raise ValueError("MORAL_COMPASS_API_BASE_URL must be set")

        session_id = session_id or os.environ.get("SESSION_ID")
        if not session_id:
            raise ValueError("SESSION_ID must be provided for single-user comprehensive test")

        logger.info("Authenticating via SESSION_ID...")
        self.auth_token = get_token_from_session(session_id)
        os.environ["JWT_AUTHORIZATION_TOKEN"] = self.auth_token
        self.username = _get_username_from_token(self.auth_token)
        log_kv("Auth Details", {"username": self.username, "token_masked": self.auth_token[:6] + "***"})

        self.client = MoralcompassApiClient(api_base_url=api_base_url, auth_token=self.auth_token)

        # Fixed table behavior:
        # Prefer TEST_TABLE_ID if provided. Otherwise, derive from TEST_PLAYGROUND_URL as <playgroundId>-mc.
        explicit_table = os.environ.get("TEST_TABLE_ID")
        pg_url = os.environ.get("TEST_PLAYGROUND_URL")

        if explicit_table and explicit_table.strip():
            self.test_table_id = explicit_table.strip()
            # If playground_url not provided, synthesize one tied to table id for metadata
            self.playground_url = pg_url or f"https://example.com/playground/{self.test_table_id}"
        else:
            # Derive stable table id from playground URL or use a shared default
            self.playground_url = pg_url or "https://example.com/playground/shared-comprehensive"
            parts = [p for p in urlparse(self.playground_url).path.split('/') if p]
            playground_id = parts[-1] if parts else "shared-comprehensive"
            self.test_table_id = f"{playground_id}-mc"

        # Random selections for this run
        random.seed()  # system entropy
        self.accuracy = random.choice(self.ACCURACY_CHOICES)
        self.tasks = 10  # base total tasks for this challenge
        self.team = random.choice(self.TEAM_CHOICES)

        # Random partial completion ratio (30–80%), then clamp to [0,1]
        raw_ratio = random.uniform(self.PARTIAL_RATIO_MIN, self.PARTIAL_RATIO_MAX)
        self.partial_ratio = max(0.0, min(1.0, raw_ratio))
        # Compute integer tasks_completed based on ratio
        self.partial_tasks_completed = max(0, min(self.tasks, int(round(self.tasks * self.partial_ratio))))

        log_kv("Test Config", {
            "test_table_id": self.test_table_id,
            "playground_url": self.playground_url,
            "selected_accuracy": self.accuracy,
            "accuracy_choices": self.ACCURACY_CHOICES,
            "total_tasks": self.tasks,
            "selected_team": self.team,
            "team_choices": self.TEAM_CHOICES,
            "partial_ratio": f"{self.partial_ratio:.2f}",
            "partial_tasks_completed": self.partial_tasks_completed,
            "skip_cleanup": SKIP_CLEANUP,
        })

        # Will store full ranking outputs for final summary
        self.last_individual_rankings: List[Dict[str, Any]] = []
        self.last_team_rankings: List[Dict[str, Any]] = []

        self.errors = []
        self.passed_tests = 0
        self.total_tests = 0

    def log_test_start(self, name):
        self.total_tests += 1
        logger.info("\n" + "=" * 70)
        logger.info(f"TEST: {name}")
        logger.info("=" * 70)

    def log_pass(self, name, msg=""):
        self.passed_tests += 1
        logger.info(f"✅ PASS: {name}")
        if msg:
            logger.info(f"   {msg}")

    def log_fail(self, name, err):
        self.errors.append(f"{name}: {err}")
        logger.error(f"❌ FAIL: {name}")
        logger.error(f"   {err}")

    def cleanup_table(self):
        logger.info("Cleaning up test table (if exists)...")
        try:
            self.client.delete_table(self.test_table_id)
            logger.info(f"Cleaned up test table: {self.test_table_id}")
        except Exception as e:
            logger.info(f"Cleanup continued (delete may be disabled or table missing): {e}")

    def ensure_table_exists(self):
        name = "Ensure Table Exists"
        self.log_test_start(name)
        try:
            # Check table
            try:
                table = self.client.get_table(self.test_table_id)
                log_kv("get_table (pre-check)", {"table_id": table.table_id, "user_count": table.user_count})
                self.log_pass(name, "Table already exists")
                return True
            except NotFoundError:
                logger.info("Table not found. Attempting to create...")
            except ApiClientError as e:
                logger.info(f"get_table error (will attempt create): {e}")

            # Create table
            create_payload = {
                "table_id": self.test_table_id,
                "display_name": f"Moral Compass Integration Test - Shared Table",
                "playground_url": self.playground_url
            }
            log_kv("create_table Request", create_payload)
            res = self.client.create_table(**create_payload)
            log_kv("create_table Response", res)

            time.sleep(0.5)
            table = self.client.get_table(self.test_table_id)
            log_kv("get_table (post-create)", {"table_id": table.table_id, "user_count": table.user_count})
            self.log_pass(name, f"Created and verified table: {self.test_table_id}")
            return True
        except Exception as e:
            self.log_fail(name, str(e))
            return False

    def test_create_user_full_completion(self):
        name = "Create Authenticated User (Full Completion)"
        self.log_test_start(name)
        try:
            request_payload = {
                "table_id": self.test_table_id,
                "username": self.username,
                "metrics": {"accuracy": self.accuracy},
                "tasks_completed": self.tasks,
                "total_tasks": self.tasks,
                "questions_correct": 0,
                "total_questions": 0,
                "primary_metric": "accuracy",
                "team_name": self.team,
            }
            log_kv("UpdateMoralCompass Request", request_payload)
            res = self.client.update_moral_compass(**request_payload)
            log_kv("UpdateMoralCompass Response", res)
            actual = float(res.get("moralCompassScore", 0))
            expected = self.accuracy  # ratio == 1
            log_kv("Score Check", {"actual": actual, "expected": expected})
            if abs(actual - expected) < 0.01:
                self.log_pass(name, f"Score correct for full completion: {actual:.4f}")
            else:
                self.log_fail(name, f"Score mismatch: {actual:.4f} (expected={expected:.4f})")
        except Exception as e:
            self.log_fail(name, str(e))

    def test_create_user_partial_completion(self):
        name = "Create/Update Authenticated User (Partial Completion)"
        self.log_test_start(name)
        try:
            request_payload = {
                "table_id": self.test_table_id,
                "username": self.username,
                "metrics": {"accuracy": self.accuracy},
                "tasks_completed": self.partial_tasks_completed,
                "total_tasks": self.tasks,
                "questions_correct": 0,
                "total_questions": 0,
                "primary_metric": "accuracy",
                "team_name": self.team,
            }
            log_kv("UpdateMoralCompass Request (Partial)", request_payload)
            res = self.client.update_moral_compass(**request_payload)
            log_kv("UpdateMoralCompass Response (Partial)", res)
            actual = float(res.get("moralCompassScore", 0))
            expected = self.accuracy * (self.partial_tasks_completed / self.tasks)
            log_kv("Score Check (Partial)", {"actual": actual, "expected": expected})
            if abs(actual - expected) < 0.01:
                self.log_pass(name, f"Score correct for partial completion: {actual:.4f}")
            else:
                self.log_fail(name, f"Score mismatch (partial): {actual:.4f} (expected={expected:.4f})")
        except Exception as e:
            self.log_fail(name, str(e))

    def list_all_users(self) -> List[Dict[str, Any]]:
        resp = self.client.list_users(table_id=self.test_table_id, limit=500)
        users = resp.get("users", [])
        log_kv("ListUsers Summary", {"count": len(users)})
        for u in users:
            logger.info(f"UserRow: username={u.get('username')} score={u.get('moralCompassScore')} team={u.get('teamName')} tasks={u.get('tasksCompleted')}/{u.get('totalTasks')} q={u.get('questionsCorrect')}/{u.get('totalQuestions')}")
        return users

    def compute_individual_rankings(self, users: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        def sort_key(x):
            return (float(x.get('moralCompassScore', 0) or 0.0), x.get('submissionCount', 0))
        ranked = sorted(users, key=sort_key, reverse=True)
        rankings: List[Dict[str, Any]] = []
        for idx, u in enumerate(ranked, 1):
            rankings.append({
                "rank": idx,
                "username": u.get("username"),
                "moralCompassScore": float(u.get("moralCompassScore", 0) or 0.0),
                "teamName": u.get("teamName"),
                "tasksCompleted": u.get("tasksCompleted"),
                "totalTasks": u.get("totalTasks"),
                "questionsCorrect": u.get("questionsCorrect"),
                "totalQuestions": u.get("totalQuestions")
            })
        return rankings

    def test_individual_ranking(self):
        name = "Individual Ranking by Moral Compass Score"
        self.log_test_start(name)
        try:
            time.sleep(0.5)
            users = self.list_all_users()
            rankings = self.compute_individual_rankings(users)
            self.last_individual_rankings = rankings

            logger.info("Current Individual Rankings (full list):")
            logger.info(json.dumps(rankings, indent=2))

            my_rank_entry = next((r for r in rankings if r["username"] == self.username), None)
            if my_rank_entry is None:
                self.log_fail(name, "Authenticated user not found in ranking list")
                return
            self.log_pass(name, f"Authenticated user's current rank: #{my_rank_entry['rank']}")
            log_kv("Individual Ranking Result", {"my_rank": my_rank_entry['rank'], "total_users": len(rankings)})
        except Exception as e:
            self.log_fail(name, str(e))

    def compute_team_rankings(self, users: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        team_scores: Dict[str, float] = {}
        team_counts: Dict[str, int] = {}
        for u in users:
            team = u.get('teamName')
            if not team:
                continue
            score = float(u.get('moralCompassScore', 0) or 0.0)
            team_scores[team] = team_scores.get(team, 0.0) + score
            team_counts[team] = team_counts.get(team, 0) + 1
        team_avgs: List[Tuple[str, float]] = [(team, team_scores[team] / team_counts[team]) for team in team_scores]
        ranked_teams = sorted(team_avgs, key=lambda kv: kv[1], reverse=True)
        rankings: List[Dict[str, Any]] = []
        for idx, (team, avg) in enumerate(ranked_teams, 1):
            rankings.append({
                "rank": idx,
                "teamName": team,
                "averageScore": avg,
                "members": team_counts.get(team, 0)
            })
        return rankings

    def test_team_ranking(self):
        name = "Team Ranking by Average Score"
        self.log_test_start(name)
        try:
            users = self.list_all_users()
            rankings = self.compute_team_rankings(users)
            self.last_team_rankings = rankings

            logger.info("Current Team Rankings (full list):")
            logger.info(json.dumps(rankings, indent=2))

            my_team = self.team
            my_team_entry = next((r for r in rankings if r["teamName"] == my_team), None)
            if my_team_entry is None:
                self.log_fail(name, f"User's team '{my_team}' not found in team rankings")
                return
            self.log_pass(name, f"User's current team rank: #{my_team_entry['rank']}")
            log_kv("Team Ranking Result", {"my_team": my_team, "my_team_rank": my_team_entry['rank'], "total_teams": len(rankings)})
        except Exception as e:
            self.log_fail(name, str(e))

    def run_all(self):
        logger.info("\n" + "=" * 80)
        logger.info("MORAL COMPASS COMPREHENSIVE INTEGRATION TEST SUITE (Single-User, Shared Table)")
        logger.info("=" * 80)
        log_kv("Run Metadata", {
            "run_id": RUN_ID,
            "started_at": datetime.utcnow().isoformat() + "Z",
            "log_file": LOG_FILE,
        })

        # Ensure table exists
        if not self.ensure_table_exists():
            logger.error("Table setup failed, aborting subsequent tests.")
        else:
            # Full completion with selected accuracy and team
            self.test_create_user_full_completion()
            # Partial completion using random 30–80% ratio
            self.test_create_user_partial_completion()
            # Individual rank from current table content (full list returned)
            self.test_individual_ranking()
            # Team rank by average score (full list returned)
            self.test_team_ranking()

        logger.info("\n" + "=" * 80)
        logger.info("TEST SUMMARY")
        logger.info("=" * 80)
        summary = {
            "Total Tests": self.total_tests,
            "Passed": self.passed_tests,
            "Failed": len(self.errors),
        }
        log_kv("Summary", summary)

        # Emit full rankings again in summary for easy artifact parsing
        logger.info("--- Full Individual Rankings (Summary) ---")
        logger.info(json.dumps(self.last_individual_rankings, indent=2))
        logger.info("--- end ---")
        logger.info("--- Full Team Rankings (Summary) ---")
        logger.info(json.dumps(self.last_team_rankings, indent=2))
        logger.info("--- end ---")

        if self.errors:
            logger.error("\nFailed Tests:")
            for e in self.errors:
                logger.error(f"  • {e}")
            return False

        logger.info("\n✅ ALL TESTS PASSED!")
        logger.info("=" * 80)
        return True

    def cleanup(self):
        logger.info("\nCleaning up test resources...")
        if SKIP_CLEANUP:
            logger.info("MC_SKIP_CLEANUP=true → skipping table deletion to preserve cumulative rankings.")
        else:
            try:
                self.cleanup_table()
            except Exception as e:
                logger.warning(f"Cleanup failed: {e}")
        logger.info(f"Logs written to: {LOG_FILE}")


def main():
    api_base_url = os.environ.get("MORAL_COMPASS_API_BASE_URL")
    if not api_base_url:
        logger.error("MORAL_COMPASS_API_BASE_URL environment variable is required")
        sys.exit(1)

    session_id = os.environ.get("SESSION_ID")
    if not session_id:
        logger.error("SESSION_ID environment variable is required")
        sys.exit(1)

    suite = MoralCompassIntegrationTest(api_base_url=api_base_url, session_id=session_id)
    ok = False
    try:
        ok = suite.run_all()
    finally:
        suite.cleanup()

    sys.exit(0 if ok else 1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Tests for Bias Detective combined score and zero-score tied-last ranking.

These tests verify that:
1. _calculate_combined_score computes correctly
2. _enforce_zero_tied_last_rank enforces last-place semantics when combined score and ethical progress are both zero
3. _compute_user_stats applies zero-score enforcement correctly
4. Team ranks are based on average accuracy (not max)

Run with: pytest tests/test_bias_detective_combined_score.py -v
"""

import pytest
from unittest.mock import MagicMock, patch
import pandas as pd

# Import functions to test
from aimodelshare.moral_compass.apps.bias_detective import (
    _calculate_combined_score,
    _enforce_zero_tied_last_rank,
    _compute_user_stats,
    _user_stats_cache,
)


def test_calculate_combined_score():
    """Test _calculate_combined_score helper function."""
    # Test basic calculation
    # combined = accuracy × (ethical_progress_pct / 100) × 100
    # ethical_progress_pct = (tasks_completed / max_points) × 100
    
    # Case 1: accuracy=0.92, tasks_completed=10, max_points=21
    # ethical_progress_pct = (10/21) * 100 = 47.619...
    # combined = 0.92 * (47.619/100) * 100 = 43.81...
    result1 = _calculate_combined_score(0.92, 10, 21)
    assert abs(result1 - 43.81) < 0.1, f"Expected ~43.81, got {result1}"
    
    # Case 2: Zero tasks completed
    result2 = _calculate_combined_score(0.92, 0, 21)
    assert result2 == 0.0, f"Expected 0.0 with zero tasks, got {result2}"
    
    # Case 3: All tasks completed
    result3 = _calculate_combined_score(0.92, 21, 21)
    assert abs(result3 - 92.0) < 0.1, f"Expected 92.0 with all tasks, got {result3}"
    
    # Case 4: Max points is zero (edge case)
    result4 = _calculate_combined_score(0.92, 5, 0)
    assert result4 == 0.0, f"Expected 0.0 with max_points=0, got {result4}"
    
    # Case 5: Tasks exceed max (should cap at 100%)
    result5 = _calculate_combined_score(0.92, 25, 21)
    assert abs(result5 - 92.0) < 0.1, f"Expected 92.0 (capped), got {result5}"


def test_enforce_zero_tied_last_rank_both_zero():
    """Test _enforce_zero_tied_last_rank when both combined and ethical progress are zero."""
    # When combined score and ethical progress are both zero, should enforce tied-last
    user_rank, team_rank = _enforce_zero_tied_last_rank(
        combined_score=0.0,
        ethical_progress_pct=0.0,
        user_rank=None,
        team_rank=None,
        total_individuals=10,
        total_teams=5
    )
    
    assert user_rank == 10, f"Expected user_rank=10 (last place), got {user_rank}"
    assert team_rank == 5, f"Expected team_rank=5 (last place), got {team_rank}"


def test_enforce_zero_tied_last_rank_non_zero_combined():
    """Test _enforce_zero_tied_last_rank when combined score is non-zero."""
    # When combined score is non-zero, should NOT enforce tied-last
    user_rank, team_rank = _enforce_zero_tied_last_rank(
        combined_score=43.81,
        ethical_progress_pct=47.6,
        user_rank=3,
        team_rank=2,
        total_individuals=10,
        total_teams=5
    )
    
    # Ranks should remain unchanged
    assert user_rank == 3, f"Expected user_rank=3 (unchanged), got {user_rank}"
    assert team_rank == 2, f"Expected team_rank=2 (unchanged), got {team_rank}"


def test_enforce_zero_tied_last_rank_non_zero_ethical():
    """Test _enforce_zero_tied_last_rank when ethical progress is non-zero."""
    # When ethical progress is non-zero, should NOT enforce tied-last
    # even if combined score is zero (can happen with zero accuracy)
    user_rank, team_rank = _enforce_zero_tied_last_rank(
        combined_score=0.0,
        ethical_progress_pct=47.6,
        user_rank=3,
        team_rank=2,
        total_individuals=10,
        total_teams=5
    )
    
    # Ranks should remain unchanged
    assert user_rank == 3, f"Expected user_rank=3 (unchanged), got {user_rank}"
    assert team_rank == 2, f"Expected team_rank=2 (unchanged), got {team_rank}"


def test_compute_user_stats_uses_average_for_team_rank():
    """Test that _compute_user_stats uses average accuracy for team ranking."""
    # Mock leaderboard data with multiple teams
    mock_df = pd.DataFrame([
        {"username": "alice", "accuracy": 0.95, "Team": "Team A"},
        {"username": "bob", "accuracy": 0.85, "Team": "Team A"},
        {"username": "charlie", "accuracy": 0.90, "Team": "Team B"},
        {"username": "dave", "accuracy": 0.80, "Team": "Team B"},
    ])
    
    # Team A average: (0.95 + 0.85) / 2 = 0.90
    # Team B average: (0.90 + 0.80) / 2 = 0.85
    # So Team A should rank #1
    
    with patch('aimodelshare.moral_compass.apps.bias_detective._fetch_leaderboard') as mock_fetch:
        mock_fetch.return_value = mock_df
        
        stats = _compute_user_stats("bob", "fake-token")
        
        # Bob is on Team A, which should rank #1
        assert stats["team_name"] == "Team A"
        assert stats["team_rank"] == 1, f"Expected team_rank=1 for Team A, got {stats['team_rank']}"


def test_compute_user_stats_applies_zero_score_enforcement():
    """Test that _compute_user_stats applies zero-score tied-last enforcement."""
    # Clear cache to avoid flakiness
    _user_stats_cache.clear()
    
    # Mock leaderboard with multiple users
    mock_df = pd.DataFrame([
        {"username": "alice", "accuracy": 0.95, "Team": "Team A"},
        {"username": "bob", "accuracy": 0.85, "Team": "Team B"},
        {"username": "charlie", "accuracy": 0.92, "Team": "Team C"},
    ])
    
    with patch('aimodelshare.moral_compass.apps.bias_detective._fetch_leaderboard') as mock_fetch:
        with patch('aimodelshare.moral_compass.apps.bias_detective.random.choice', return_value="Team B"):
            mock_fetch.return_value = mock_df
            
            # For a new user with 0 tasks completed, the zero-score guard should apply
            # in _compute_user_stats since it assumes tasks_completed=0
            stats = _compute_user_stats("bob", "fake-token")
            
            # With zero tasks completed, combined score = 0.85 * 0 * 100 = 0
            # Zero-score guard should set rank to last place (total_individuals)
            assert stats["rank"] == stats["total_individuals"], \
                f"Expected rank={stats['total_individuals']} (tied-last) for zero tasks, got {stats['rank']}"
            assert stats["total_individuals"] == 3
            assert stats["total_teams"] == 3


def test_compute_user_stats_counts_total_individuals_and_teams():
    """Test that _compute_user_stats correctly counts total individuals and teams."""
    # Mock leaderboard with specific counts
    mock_df = pd.DataFrame([
        {"username": "user1", "accuracy": 0.95, "Team": "Team A"},
        {"username": "user2", "accuracy": 0.90, "Team": "Team A"},
        {"username": "user3", "accuracy": 0.85, "Team": "Team B"},
        {"username": "user4", "accuracy": 0.80, "Team": "Team C"},
        {"username": "user5", "accuracy": 0.75, "Team": "Team C"},
    ])
    
    with patch('aimodelshare.moral_compass.apps.bias_detective._fetch_leaderboard') as mock_fetch:
        mock_fetch.return_value = mock_df
        
        stats = _compute_user_stats("user3", "fake-token")
        
        assert stats["total_individuals"] == 5, f"Expected 5 individuals, got {stats['total_individuals']}"
        assert stats["total_teams"] == 3, f"Expected 3 teams, got {stats['total_teams']}"


if __name__ == '__main__':
    pytest.main([__file__, '-v'])
